/*
* Copyright (c) 2007, Jakub Trznadel
* mailto: trznadelj@o2.pl
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* Generator of an image allowing generation of PAL signal from simple vga card */

#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

/* Yeah, include image generated by GIMP! */
#include "PICT0005.h"

/* VGA image defines, as typed into Xorg.conf:                              */
/* Modeline  "MaxTV"       17.734475   1125 1125 1134 1135  623 623 624 625 */
#define RES_X                   1128
#define RES_Y                    625
#define DOT_CLOCK_FREQ      17734475 /*  4*PAL */

/* RGB signals from VGA are 0.7V p-p */
#define MAX_V                    0.7 /*  700mV */

/* PAL image defines */
#define PAL_LINES                625
#define PAL_COLOR_FREQ       4429712 /*  >4MHz */

/* Color support (not tested!) */
#define PAL_BURST_LENGTH_IN_DOT_CLOCKS  (int) ((DOT_CLOCK_FREQ/PAL_COLOR_FREQ)*12)
#define FREQ_RATIO_RAD         2*M_PI*PAL_COLOR_FREQ/DOT_CLOCK_FREQ /* radians */
#define FREQ_RATIO_SIN         sin(FREQ_RATIO_RAD)
#define FREQ_RATIO_COS         cos(FREQ_RATIO_RAD)
#define PAL_BURST_AMPLITUDE_V  0.15 /* 150mV */
#define PAL_BURST_AMPLITUDE    255*PAL_BURST_AMPLITUDE_V/MAX_V

/* Voltage levels for VGA */
#define PAL_BLANK                  0
#define PAL_BLACK                109 /* 0.3/0.7 * 255 = 109 */
#define PAL_WHITE                255

/* Scale 0-255 grayscale image to accessible levels */
#define SCALE_BW(x) (uint8) ( (((unsigned int) x)* (PAL_WHITE-PAL_BLACK))/(PAL_WHITE)  + PAL_BLACK)

/* Some handfull macros.. */
#define uint8  unsigned char
#define uint16 unsigned short
#define uint32 unsigned int
#define PUT_B(b)  { uint8  temp_var=b; write( fd, &temp_var, 1); }
#define PUT_2B(b) { uint16 temp_var=b; write( fd, &temp_var, 2); }
#define PUT_4B(b) { uint32 temp_var=b; write( fd, &temp_var, 4); }
#define CP  printf("[%i]\n", __LINE__);


typedef struct 
{ 
    uint16 res_x, res_y;
    uint8  pict[RES_Y][RES_X];
} image_t;


void image_clear( image_t* thiz)
{
    memset( thiz->pict, 0, thiz->res_x * thiz->res_y);
}

/* return array index from coordinates: PAL line and x */
int  image_get_pos( image_t *thiz, int pal_line, int pal_x)
{
     long long z1, z2;

     z1 = (thiz->res_x  * thiz->res_y)*pal_line;
     z2 = (pal_x        * thiz->res_x);
     z1*=RES_X;
     z2*=PAL_LINES;

     return  (z1 + z2)/(PAL_LINES*RES_X);
}

/* simple memset wrapper */
void image_setline( image_t* thiz, int pos0, int pos1, uint8 val)
{
    int blen= thiz->res_x*thiz->res_y;

    if ( pos1 >= blen ) pos1 = blen-1;    
    if ( pos0 >= blen ) pos0 = blen-1;    
    if ( pos0 >= pos1 ) exit(0);    
    memset( (char*) &(thiz->pict[0][0]) + pos0, val, pos1-pos0);
}

/* also simple memset wrapper, BUT us0 and us1 are in microseconds */
void image_set_line_backtraces( image_t* thiz, int pal_line, int us0, int us1, uint8 val)
{
    image_setline( thiz, image_get_pos( thiz, pal_line, us0*thiz->res_x/64), 
                         image_get_pos( thiz, pal_line, us1*thiz->res_x/64), val);
}

/* Generator for lines in vertical retrace - 4 types of lines */
void image_pal_line_vertical( image_t *thiz, int line, uint8 type)
{
    switch( type )
    {
	case 0: // 0 1
	    image_set_line_backtraces( thiz, line,  0, 64, PAL_BLANK);
	    image_set_line_backtraces( thiz, line, 28, 32, PAL_BLACK);
	    image_set_line_backtraces( thiz, line, 60, 64, PAL_BLACK);
 	break;
	case 1: // 2
	    image_set_line_backtraces( thiz, line,  0, 64, PAL_BLANK);
	    image_set_line_backtraces( thiz, line, 28, 32, PAL_BLACK);
	    image_set_line_backtraces( thiz, line, 36, 64, PAL_BLACK);

	break;
	case 2:
	    image_set_line_backtraces( thiz, line,  0, 64, PAL_BLANK);
	    image_set_line_backtraces( thiz, line,  0,  4, PAL_BLACK);
	    image_set_line_backtraces( thiz, line, 32, 36, PAL_BLACK);
	break;
	case 3:
	    image_set_line_backtraces( thiz, line,  0, 64, PAL_BLANK);
	    image_set_line_backtraces( thiz, line,  4, 32, PAL_BLACK);
	    image_set_line_backtraces( thiz, line, 60, 64, PAL_BLACK);
	break;
    }
}

/* Add color burst to the line, not proved that it's working */
void image_set_colorburst( image_t* thiz, int pal_line)
{
    int addr = image_get_pos( thiz, pal_line, 6*RES_X/64 );
    int i;
    float f_val, f_val1, f_val2;

    for( i=0; i< PAL_BURST_LENGTH_IN_DOT_CLOCKS; i++)
    {
	if (0==i)
	{
	    f_val = (pal_line&1);
	}
	else if (1==i)
	{
	    f_val = (pal_line&1)?FREQ_RATIO_COS:FREQ_RATIO_SIN;
	}
	else
	{
	    f_val = 2*FREQ_RATIO_COS*f_val1-f_val2;
	}	
	/* calculate sine through Z-transform */
	((char*) (&(thiz->pict[0][0])))[addr+i] = f_val*PAL_BURST_AMPLITUDE + PAL_BLACK;
        f_val2=f_val1;	
        f_val1=f_val;
    }
}


/* I like simple functions. */
void put_pixel( image_t* thiz, int x, int y, uint8 val)
{
    int pos;

    if (y>310) return;
    if (y<0) return; 
    if (x<=thiz->res_x*3/16) return; /* first 12uS are not used by image */
    if (x>=thiz->res_x) return;

    /* and now write to both first and second page of PAL frame. */
    pos = image_get_pos( thiz, y+5, x);
    memset( (char*) &(thiz->pict[0][0]) + pos, val, 1);
    pos = image_get_pos( thiz, y+320, x);
    memset( (char*) &(thiz->pict[0][0]) + pos, val, 1);
}

void image_create_pal_frame( image_t* thiz)
{
    int i,j;

    /* Vertical retrace signal */
    image_pal_line_vertical( thiz, 622, 2);
    image_pal_line_vertical( thiz, 623, 2);
    image_pal_line_vertical( thiz, 624, 2);
    image_pal_line_vertical( thiz,   0, 0);
    image_pal_line_vertical( thiz,   1, 0);
    image_pal_line_vertical( thiz,   2, 1);
    image_pal_line_vertical( thiz,   3, 2);
    image_pal_line_vertical( thiz,   4, 2);
    image_pal_line_vertical( thiz, 310, 2);
    image_pal_line_vertical( thiz, 311, 2);
    image_pal_line_vertical( thiz, 312, 3);
    image_pal_line_vertical( thiz, 313, 0);
    image_pal_line_vertical( thiz, 314, 0);
    image_pal_line_vertical( thiz, 315, 2);
    image_pal_line_vertical( thiz, 316, 2);

    /* normal PAL line */
    for( i=5; i<621; i++)
    {
	if (i==310) i=317; /* skip vertical ratrace	 */

        /* vertical SYNC pulse */
        image_set_line_backtraces( thiz, i,  0,  4, PAL_BLANK);
        /* rest of line is empty */
        image_set_line_backtraces( thiz, i,  4, 64, PAL_BLACK);
        /* add color burst signal */
	image_set_colorburst     ( thiz, i);
    }

    /* Add a picture*/
    for( i=0; i<width; i++)
       for( j=0; j<height; j++)    {
	   put_pixel( thiz, 2*i+250, 250-j, SCALE_BW(header_data[i+j*width])  );
	   put_pixel( thiz, 2*i+251, 250-j, SCALE_BW(header_data[i+j*width])  );
       }

    /* color testing */
    for( i=0; i<400; i++)
	for( j=0; j<20; j++)
           put_pixel( thiz, 250+i, 25+j, 200-(i/10)+4*j*sin(2*M_PI*i/4));       
}

/* Write result array to BMP file. It's simple enought :) */
void image_write_bmp( image_t* thiz, char * filename)
{
    int fd = open ( filename, O_WRONLY | O_CREAT);
    int i;

    if ( -1 == fd )
    {
	printf("error opening %s for write!\n", filename);
	exit(0);
    }

    PUT_B('B');    
    PUT_B('M');
    PUT_4B(1078 + thiz->res_x * thiz->res_y); //file size
    PUT_4B(0); //res1 res2
    PUT_4B(1078); 
    PUT_4B(40); 
    PUT_4B(thiz->res_x); 
    PUT_4B(thiz->res_y); 
    PUT_2B(1); //planes
    PUT_2B(8); //bits per pixel
    PUT_4B(0); //compression
    PUT_4B(0); //data size
    PUT_4B(0); //
    PUT_4B(0); //
    PUT_4B(0); //
    PUT_4B(0); //

    for( i=0; i<256; i++)
    {
	PUT_B(i); PUT_B(i); PUT_B(i); PUT_B(0); //r g b 0	
    }
    write( fd, &(thiz->pict[0][0]), thiz->res_x * thiz->res_y);
    close(fd);
    printf("Image size: %ix%i is in file %s\n", thiz->res_x, thiz->res_y, filename);
}


int main()
{
    image_t my_image;

    my_image.res_x = RES_X;
    my_image.res_y = RES_Y;

    image_clear           ( &my_image );
    image_create_pal_frame( &my_image );
    image_write_bmp       ( &my_image, "first.bmp");    

    return 0;
}
